import numpy as np
from ..core.Reader import Reader
import datetime

class AQ500(Reader):

    def __init__(self):
        super().__init__(False)

    @staticmethod
    def util_process_time(timestring):
        hour = timestring[:2]
        minute = timestring[3:5]
        seconds = timestring[6:11]
        time = int(hour) * 3600
        time += int(minute) * 60
        time += float(seconds)
        return time

    @staticmethod
    def str_to_num(string1):
        try:
            return int(string1)
        except ValueError:
            try:
                return float(string1)
            except ValueError:
                return string1

    def correct_ws(ws_array,range_array):
        corr_array = np.ones_like(ws_array)
        
        bool_mask_80=(range_array <=80)
        corr_array[:,bool_mask_80]=0.98
        
        bool_mask_140=(range_array >= 140)
        corr_array[:,bool_mask_140]=0.97

        bool_mask_between = (~bool_mask_80) & (~bool_mask_140)
        corr_array[:,bool_mask_between] = np.tile(range_array[bool_mask_between]*(-1/6000)+(149/150), (len(ws_array),1) )

        return ws_array * corr_array

    def accepts_file(self, filename):
        return filename.endswith('.txt')

    def output_filename(self, filename):
        return filename[:-4]

    def read_to(self, output_dataset, input_filepath, configs, appending):

        # read file
        with open(input_filepath, encoding='latin-1') as f:
            data = f.readlines()
            data = [line.strip() for line in data]
            temp_headerlength = data.index('[EOH]')
            parameters = {line.split('=')[0]: line.split('=')[1].strip() for line in data[:temp_headerlength] if ('=' in line)}
            parameters['HeaderLength'] = temp_headerlength
            parameters['Measurement heights'] = np.arange(int(parameters['Lowest level(LL m)']),int(parameters['Highest level(HL m)'])+1,int(parameters['Interval(m)']) )

        	#create datafield dictionary  {column_name:np.array(columns)}
            datafield = {line.split(':')[1].strip() : np.array((line.split(':')[0][11:]).split(','),dtype=int)-2 for line in data[:temp_headerlength] if (':' in line)}
            
            # create the dimensions
            output_dataset.createDimension('range', len(parameters['Measurement heights']))
            output_dataset.createDimension('time', None)

            # create the coordinate variables
            range1 = output_dataset.createVariable('range', 'f4', ('range',))
            range1.units = 'm'
            range1.long_name = 'range_gate_distance_from_lidar'
            range1[:] = np.array(parameters['Measurement heights'])

            time = output_dataset.createVariable('time', str, ('time',))
            time.units = 's'
            time.long_name = 'Time UTC in ISO 8601 format yyyy-mm-ddThh:mm:ssZ'

            # create the data variables
            scan_type = output_dataset.createVariable('scan_type', 'i')
            scan_type.units = 'none'
            scan_type.long_name = 'scan_type_of_the_measurement'
            scan_type[:] = 1

            accumulation_time = output_dataset.createVariable('accumulation_time', 'f4')
            accumulation_time.units = 'seconds'
            accumulation_time.long_name = 'time_for_spectral_accumulation'
            accumulation_time[:] = 1.0

            # create the measurement variables
            T_external = output_dataset.createVariable('T_external', 'f4', ('time',))
            T_external.units = 'degrees C'
            T_external.long_name = 'temperature'

            rh = output_dataset.createVariable('rh', 'f4', ('time',))
            rh.units = 'degrees'
            rh.long_name = 'lidar_yaw_angle'
            
            p = output_dataset.createVariable('p', 'f4', ('time',))
            p.units = 'degrees'
            p.long_name = 'lidar_yaw_angle'

            WS = output_dataset.createVariable('WS', 'f4', ('time', 'range'))
            WS.units = 'm.s-1'
            WS.long_name = 'mean of scalar wind speed'
            
            DIR = output_dataset.createVariable('DIR', 'f4', ('time', 'range'))
            DIR.units = 'degrees north'
            DIR.long_name = 'wind direction from north'
            
            signal_quality = output_dataset.createVariable('signal_quality', 'f4', ('time', 'range'))
            signal_quality.units = 'percent'
            signal_quality.long_name = 'signal quality'

            # create np.array of dataset without timestamp and cast to float
            data_timeseries = [line[:-1].split(',') for line in data[temp_headerlength+2:-2]]
            data_timeseries_array = np.array(data_timeseries)[:,1:].astype(float)
            
            timestamp_input = [datetime.datetime.strptime(row[0],'%Y%m%d %H:%M') for row in data_timeseries]
            timestamp_iso8601 = [value.isoformat()+'Z' for value in timestamp_input]
            output_dataset.variables['time'][:] = np.array(timestamp_iso8601)

            output_dataset.variables['T_external'][:] = data_timeseries_array[:,datafield['Temperature sensor(deg C * 10)=True']]
            output_dataset.variables['rh'][:] = data_timeseries_array[:,datafield['Humidity sensor(%RH)=True']]
            output_dataset.variables['p'][:] = data_timeseries_array[:,datafield['Pressure sensor(Hp)=False']]

            # e.g. radial velocity starts at 5th column and is then repeated every 9th column
            output_dataset.variables['WS'][:, :] = AQ500.correct_ws(data_timeseries_array[:,datafield['Speed m/s(LL to HL)']], parameters['Measurement heights'])
            output_dataset.variables['DIR'][:, :] = data_timeseries_array[:,datafield['Dir degrees(LL to HL)']]
            
            # there is an error reading signal_quality for our data because there are only 30 columns, but 31 heights
            # we commented it out but it might be useful for future measurements
            # output_dataset.variables['signal_quality'][:, :] = data_timeseries_array[:,datafield['Quality(S/N*10)(LL to HL)']]


import collections


def common_iterable(obj):
    """
    Offers a single iterator for a dictionary or list.
    Based on : to https://stackoverflow.com/questions/12325608/iterate-over-a-dict-or-list-in-python
    :param obj: dictionary or list
    :return: iterator
    """
    if isinstance(obj, dict):
        return ((value, obj[value], index) for index, value in enumerate(obj.keys()))
    else:
        return ((index, value, index) for index, value in enumerate(obj))


def dict_merge(dct, merge_dct):
    """ Recursive dict merge. Inspired by :meth:``dict.update()``, instead of
    updating only top-level keys, dict_merge recurses down into dicts nested
    to an arbitrary depth, updating keys. The ``merge_dct`` is merged into
    ``dct``.
    From: https://gist.github.com/angstwad/bf22d1822c38a92ec0a9 (adapted python 3.6)
    :param dct: dict onto which the merge is executed
    :param merge_dct: dct merged into dct
    :return: None
    """
    for k, v in merge_dct.items():
        if (k in dct and isinstance(dct[k], dict)
                and isinstance(merge_dct[k], collections.Mapping)):
            dict_merge(dct[k], merge_dct[k])
        else:
            dct[k] = merge_dct[k]


def map_recursively(dct, value):
    mapped_dct = {}
    for k, v in dct.items():
        if k in dct and isinstance(dct[k], dict):
            mapped_dct[k] = map_recursively(dct[k], value)
        else:
            mapped_dct[k] = value
    return mapped_dct


def is_str(s):
    """
    Checks if a variable if a string
    :param s: a string
    :return: boolean
    """
    return isinstance(s, str)


def to_dict(*kwargs):
    print(kwargs)
    for key, value in kwargs:
        print(key, value)



import sys


class Logger:
    """
    Handles output printing.
    """

    separator = 45 * '='
    messages = {
        'separator': separator,
        'header': "{} \nWindsOf - NetCDF lidar data converters - v0.1 \n{}".format(separator, separator),
        'bad_config_path': 'No configuration file ({}) found.',
        'inp_path_missing': 'Missing input path config. Use -D to quickly set it.',
        'inp_format_missing': 'Missing input format config. Use -I to quickly set it.',
        'out_format_missing': 'Missing output format config. Use -O to quickly set it.',
        'bad_inp_format': 'Failed to load reader "{}". Native error: {}',
        'bad_out_format': 'Failed to load writer "{}". Native error: {}',
        'input_format_detected': 'Input format detected: {}.',
        'output_format_detected': 'Output format detected: {}.',
        'searching_in_path': 'Looking for input files in {}',
        'found': 'Found {}.',
        'started_r_files': 'Processing {} ...',
        'grouping': 'Grouping files...',
        'writing_file': 'Writing to {} {}.',
        'exit_msg': 'Failed.',
        'files_not_found': 'No valid files were found.',
        'loading_config': 'Loading configurations from {} .',
        'bad_config_file': 'Failed to load config file. ',
        'bad_config_formatting': 'Failed loading; {}',
        'missing_reader_param': 'The config {}, required by the "{}" reader is not set. ' +
                                'Set it under in the .yaml files.',
        'done': 'Done.',
        'about': ''
                 + '   _ _     _                 \n'
                 + '|    | (_)   | |                \n'
                 + '|    | |_  __| | __ _  ___ ___  \n'
                 + '|    | | |/ _` |/ _` |/ __/ _ \ \n'
                 + '|    | | | (_| | (_| | (_| (_) |\n'
                 + '|    |_|_|\__,_|\__,_|\___\___/  v0.0.1\n'
    }
    verbose = False
    _debug = False

    @staticmethod
    def set_args(args):
        """
        Used to set Logger arguments
        :param args: should contain 'verbose', 'debug' attributes
        :return:
        """
        Logger.verbose = args.verbose
        Logger._debug = args.debug

    @staticmethod
    def __print_std_output(prefix, msg_name, *args):
        """
        Prints to standard output.
        :param prefix: message prefix e.g [Error], [Warning]
        :param msg_name: message key
        :param args: arguments to format the message with
        :return: void
        """
        formatted_msg = args[0] if msg_name is None else Logger.messages[msg_name].format(*args)
        if prefix is None:
            print(formatted_msg)
        else:
            print(prefix, formatted_msg)

    @staticmethod
    def header():
        """
        Prints the converter header.
        :return: void
        """
        Logger.__print_std_output(None, 'header')

    @staticmethod
    def log(msg_name, *args):
        """
        Prints a message
        :param msg_name: message name
        :param args: message arguments
        :return:
        """
        Logger.__print_std_output('| ', msg_name, *args)

    @staticmethod
    def info(msg_name, *args):
        """
        Prints a message if verbose mode is set.
        :param msg_name: message name
        :param args: message arguments
        :return:
        """
        if Logger.verbose:
            Logger.__print_std_output('[Info] ', msg_name, *args)

    @staticmethod
    def debug(msg_name, *args):
        """
        Prints a message if debug mode is set.
        :param msg_name: message name
        :param args: message arguments
        :return:
        """
        if Logger._debug:
            Logger.__print_std_output('[Debug] \n', msg_name, *args)

    @staticmethod
    def warn(msg_name, *args):
        """
        Prints a warning.
        :param msg_name: message name
        :param args: message arguments
        :return:
        """
        Logger.__print_std_output('[Warn] ', msg_name, *args)

    @staticmethod
    def error(msg_name, *args):
        """
        Prints an error message and ends the program.
        :param msg_name: message name
        :param args: message arguments
        :return:
        """
        Logger.__print_std_output('[Error] ', msg_name, *args)
        sys.exit()



from lxml.etree import Element, ElementTree

import netCDF4 as nc

from ..core.Writer import Writer

NS = "http://www.unidata.ucar.edu/namespaces/netcdf/ncml-2.2"
PREFIX = '{' + NS + '}'
NS_MAP = {'ncml': NS}


class NcML(Writer):
    nc_dataset = None

    def __init__(self, dir_path, name):
        super().__init__(dir_path, name)

    def filename(self):
        return self.name + '.ncml'

    def __enter__(self):
        if not self.append:
            self.nc_dataset = nc.Dataset(self.file_path(), 'w', diskless=True)
            self.dataset = ElementTree(Element(PREFIX + "netcdf", nsmap=NS_MAP))

        return self.nc_dataset.__enter__()
    def __exit__(self, type, value, traceback):
        if not self.append:
            # Writing dimensions
            # asdasd
            for name, dim in self.nc_dataset.dimensions.items():
                element = Element(PREFIX + 'dimension')
                element.set('name', name)
                element.set('length', str(dim.size))
                if dim.isunlimited:
                    element.set('isUnlimited', "true")
                self.dataset.getroot().append(element)

            # Writing attributes
            for ncattr in self.nc_dataset.ncattrs():
                element = Element(PREFIX + 'attribute')
                element.set('name', ncattr)
                element.set('value', self.nc_dataset.getncattr(ncattr))
                self.dataset.getroot().append(element)

            # Writing variables
            for name, var in self.nc_dataset.variables.items():

                element = Element(PREFIX + 'variable')
                element.set('name', name)
                element.set('shape', " ".join(var.dimensions))
                element.set('type', var.dtype.str[1:])

                # Writing variable attributes
                for ncattr in var.ncattrs():
                    attr_elem = Element(PREFIX + 'attribute')
                    attr_elem.set('name', ncattr)
                    attr_elem.set('value', var.getncattr(ncattr))
                    element.append(attr_elem)

                if var.chunking() != 'contiguous':
                    attr_elem = Element(PREFIX + 'attribute')
                    attr_elem.set('name', '_ChunkSizes')
                    attr_elem.set('type', 'int')
                    attr_elem.set('value', " ".join([str(s) for s in var.chunking()]))
                    element.append(attr_elem)

                    self.dataset.getroot().append(element)

            return self.dataset.write(self.file_path(), xml_declaration=True, encoding="UTF-8", pretty_print=True)



from ..core.Reader import Reader
from datetime import datetime, timedelta
import numpy as np


class Windscanner(Reader):

    def __init__(self):
        super().__init__(False)

    def accepts_file(self, filename):
        return filename.endswith('wind.txt') & (len(filename) > 14)

    def output_filename(self, timestamp):
        return timestamp[:-9]

    def read_to(self, output_dataset, input_filepaths, parameters, appending):
        wind_file = input_filepaths

        with open(wind_file) as f:
            wind_file_data = f.readlines()

        wind_file_data = list(zip(*[row.strip().split(';') for row in wind_file_data]))

        if not appending:

            index_columns = 4 - (len(wind_file_data) % 4)
            range_list = [float(row[0]) for row in wind_file_data[index_columns + 4::4]]

            # create the dimensions
            output_dataset.createDimension('range', len(range_list))
            output_dataset.createDimension('time', None)

            # create the coordinate variables

            # range
            range1 = output_dataset.createVariable('range', 'f4', ('range',))
            range1.units = 'm'
            range1.long_name = 'range_gate_distance_from_lidar'
            range1[:] = range_list
            range1.comment = ''

            # time
            time = output_dataset.createVariable('time', str, ('time',))
            time.units = 's'
            time.long_name = 'Time UTC in ISO 8601 format yyyy-mm-ddThh:mm:ssZ'
            time.comment = ''

            # create the data variables
            scan_type = output_dataset.createVariable('scan_type', 'i')
            scan_type.units = 'none'
            scan_type.long_name = 'scan_type_of_the_measurement'


            # create the measurement variables VEL, CNR, WIDTH
            VEL = output_dataset.createVariable('VEL', 'f4', ('time', 'range'))
            VEL.units = 'm.s-1'
            VEL.long_name = 'radial velocity'
            VEL.comment = ''
            VEL.accuracy = ''
            VEL.accuracy_info = ''

            CNR = output_dataset.createVariable('CNR', 'f4', ('time', 'range'))
            CNR.units = 'dB'
            CNR.long_name = 'carrier-to-noise ratio'
            CNR.comment = ''
            CNR.accuracy = ''
            CNR.accuracy_info = ''

            WIDTH = output_dataset.createVariable('WIDTH', 'f4', ('time', 'range'))
            WIDTH.units = 'm.s-1'
            WIDTH.long_name = 'doppler spectrum width'
            WIDTH.comment = ''
            WIDTH.accuracy = ''
            WIDTH.accuracy_info = ''


            #%% get timestamps in ISO 8601 format
            start_date = datetime(1904,1,1)
            timestamp_seconds = [int(float(value.strip())) for value in wind_file_data[index_columns]]
            timestamp_iso8601 = [ (start_date+timedelta(seconds=value)).isoformat()+'Z' for value in timestamp_seconds]
            output_dataset.variables['time'][:] = np.array(timestamp_iso8601)
            
            #%% calculate azimuth and elevation sweeps
            azimuth_angle_temp = [float(value) for value in wind_file_data[6]]
            elevation_angle_temp = [float(value) for value in wind_file_data[7]]
            azimuth_sweep_temp = np.insert(np.abs(np.diff(azimuth_angle_temp)),0,np.nan)
            elevation_sweep_temp = np.insert(np.abs(np.diff(elevation_angle_temp)),0,np.nan)

            #%% check if a sweep is existing and writing data accordingly
            changing_azimuth = False
            changing_elevation = False
            beam_sweeping = (parameters['attributes']['beam_sweeping'] == 'true')
            
            if np.nanmedian(azimuth_sweep_temp) > 0:
                changing_azimuth = True
                azimuth_angle = output_dataset.createVariable('azimuth_angle', 'f4', ('time'))
            
                azimuth_sweep = output_dataset.createVariable('azimuth_sweep', 'f4', ('time'))
                azimuth_sweep.units = 'degrees'
                azimuth_sweep.long_name = 'azimuth_sector_swept_during_accumulation'
                azimuth_sweep.comment = ''
                azimuth_sweep.accuracy = ''
                azimuth_sweep.accuracy_info = ''
            
                output_dataset.variables['azimuth_angle'][:] = azimuth_angle_temp
                output_dataset.variables['azimuth_sweep'][:] = azimuth_sweep_temp
                
            else:
                azimuth_angle = output_dataset.createVariable('azimuth_angle', 'f4')
                output_dataset.variables['azimuth_angle'][:] = np.nanmedian(azimuth_angle_temp)
            
            
            if np.nanmedian(elevation_sweep_temp) > 0:
                changing_elevation = True
                elevation_angle = output_dataset.createVariable('elevation_angle', 'f4', ('time'))
        
                elevation_sweep = output_dataset.createVariable('elevation_sweep', 'f4', ('time'))
                elevation_sweep.units = 'degrees'
                elevation_sweep.long_name = 'elevation_sector_swept_during_accumulation'
                elevation_sweep.comment = 'Elevation sweeping from approximately 0 to 15 degrees.'
                elevation_sweep.accuracy = ''
                elevation_sweep.accuracy_info = ''
        
                output_dataset.variables['elevation_angle'][:] = elevation_angle_temp
                output_dataset.variables['elevation_sweep'][:] = elevation_sweep_temp
            else:
                elevation_angle = output_dataset.createVariable('elevation_angle', 'f4')
                output_dataset.variables['elevation_angle'][:] = np.nanmedian(elevation_angle_temp)
            
            azimuth_angle.units = 'degrees'
            azimuth_angle.long_name = 'azimuth_angle_of_lidar beam'
            azimuth_angle.comment = ''
            azimuth_angle.accuracy = ''
            azimuth_angle.accuracy_info = ''
            
            elevation_angle.units = 'degrees'
            elevation_angle.long_name = 'elevation_angle_of_lidar beam'
            elevation_angle.comment = ''
            elevation_angle.accuracy = ''
            elevation_angle.accuracy_info = ''

            #%% setting scan_type according to sweeps            
            if (not changing_azimuth) & (not changing_elevation): #case LOS
                scan_type[:] = 1
            elif (changing_azimuth) & (not changing_elevation) & (not beam_sweeping): #case DBS
                scan_type[:] = 2
            elif (changing_azimuth) & (not changing_elevation) & (beam_sweeping): #case PPI
                scan_type[:] = 4
            elif (not changing_azimuth) & (changing_elevation) & (beam_sweeping): #case RHI
                scan_type[:] = 5
            else: #case other
                scan_type[:] = 0
            
         
            #%% read vel, width, cnr out of dataset
            # e.g. radial velocity starts at 5th column and is then repeated every 9th column
            output_dataset.variables['VEL'][:, :] = list(
                zip(*[[float(value) for value in row] for row in wind_file_data[index_columns + 5::4]]))
            output_dataset.variables['WIDTH'][:, :] = list(
                zip(*[[float(value) for value in row] for row in wind_file_data[index_columns + 7::4]]))
            output_dataset.variables['CNR'][:, :] = list(
                zip(*[[float(value) for value in row] for row in wind_file_data[index_columns + 6::4]]))
            
        #%% case appending
        else: 
            index_columns = 4 - (len(wind_file_data) % 4)
            ntime = len(output_dataset.dimensions["time"])
            
            start_date = datetime(1904,1,1)
            timestamp_seconds = [int(float(value.strip())) for value in wind_file_data[index_columns]]
            timestamp_iso8601 = [ (start_date+timedelta(seconds=value)).isoformat()+'Z' for value in timestamp_seconds]
            output_dataset.variables['time'][ntime:] = np.array(timestamp_iso8601)
            
            if 'azimuth_sweep' in output_dataset.variables :
                azimuth_angle_temp = [float(value) for value in wind_file_data[6]]
                azimuth_sweep_temp = np.insert(np.abs(np.diff(azimuth_angle_temp)),0,np.nan)
                output_dataset.variables['azimuth_angle'][ntime:] = azimuth_angle_temp
                output_dataset.variables['azimuth_sweep'][ntime:] = azimuth_sweep_temp

            if 'elevation_sweep' in output_dataset.variables :
                elevation_angle_temp = [float(value) for value in wind_file_data[7]]
                elevation_sweep_temp = np.insert(np.abs(np.diff(elevation_angle_temp)),0,np.nan)
                output_dataset.variables['elevation_angle'][ntime:] = elevation_angle_temp
                output_dataset.variables['elevation_sweep'][ntime:] = elevation_sweep_temp


            # e.g. radial velocity starts at 5th column and is then repeated every 9th column
            output_dataset.variables['VEL'][ntime:, :] = list(
                zip(*[[float(value) for value in row] for row in wind_file_data[index_columns + 5::4]]))
            output_dataset.variables['WIDTH'][ntime:, :] = list(
                zip(*[[float(value) for value in row] for row in wind_file_data[index_columns + 7::4]]))
            output_dataset.variables['CNR'][ntime:, :] = list(
                zip(*[[float(value) for value in row] for row in wind_file_data[index_columns + 6::4]]))



from os import path
import pathlib

from lidaco.core.Writer import Writer

from lidaco.core.Reader import Reader

from ..common.Utils import is_str
from ..common.Logger import Logger
from .ModuleLoader import ModuleLoader
from .Config import Config


class Builder:
    """
    Main class. Orchestrates all the conversion steps.

    """

    module_loader = ModuleLoader()
    logger = None
    input_dir_path = None
    configs = {}
    args = {}

    def __init__(self,
                 config_file=None,
                 input_path=None,
                 output_format=None,
                 input_format=None,
                 context='',
                 ):
        """
        Initialization block. Loads a main config.yaml file, a reader, a writer and the remaining
        "meta-data" configurations. Overrides main configurations with the terminal arguments.
        :param context: this executable path
        :param args: terminal arguments
        :return: void
        """

        absolute_path = path.join(context, config_file)
        import_dir_path = path.dirname(absolute_path)
        import_filename = path.basename(absolute_path)

        root_configs = {
            'imports': [
                import_filename
            ],
            'parameters': {
                'input': {},
                'output': {}
            },
        }

        if input_path is not None:
            root_configs['parameters']['input']['path'] = input_path

        if input_format is not None:
            root_configs['parameters']['input']['format'] = input_format

        if output_format is not None:
            root_configs['parameters']['output']['format'] = output_format

        self.configs = Config(import_dir_path, configs=root_configs)

        if self.configs.get('parameters', 'input', 'path') is None:
            Logger.error('inp_path_missing')

        if self.configs.get('parameters', 'output', 'path') is None:
            # setting the default output folder
            # this could be generalized and encapsulated as a setter on the config class
            self.configs.configs = 'output'
            self.configs.config_paths = context

        reader = self.params('input', 'format')
        if not is_str(reader) and issubclass(reader, Reader):
            self.module_loader.set_reader(reader)
        else:
            try:
                self.module_loader.load_reader(reader)
                Logger.info('input_format_detected', self.params('input', 'format'))
            except KeyError as e:
                Logger.debug(e)
                Logger.error('inp_format_missing')
            except Exception as e:
                Logger.debug(e)
                Logger.error('bad_inp_format', self.params('input', 'format'), str(e))

        writer = self.params('output', 'format')
        if not is_str(writer) and issubclass(writer, Writer):
            self.module_loader.set_writer(writer)
        else:
            try:
                self.module_loader.load_writer(writer)
                Logger.info('output_format_detected', self.params('output', 'format'))
            except KeyError as e:
                Logger.debug(e)
                Logger.error('out_format_missing')
            except Exception as e:
                Logger.debug(e)
                Logger.error('bad_out_format', self.params('output', 'format'), str(e))

    def params(self, *keys):
        return self.configs.get('parameters', *keys)

    def read_attributes(self, dataset):
        """
        Reads attributes into the dataset. The attributes should be specified in the .yaml configuration
        files under 'attributes:' key.
        :param dataset: a core data model / netcdf4 dataset object.
        :return: void
        """
        if 'attributes' in self.configs:
            for key, value in self.configs['attributes'].items():
                setattr(dataset, key, value)
                
    def read_variables(self, dataset):
        """
        Reads variables into the dataset. The variables should be specified in the .yaml configuration
        files under 'variables:' key.
        :param dataset: a core data model / netcdf4 dataset object.
        :return: void
        """
        if 'variables' in self.configs:
            for variable_name, variable_dict in self.configs['variables'].items():
                if variable_name not in dataset.variables:
                    temp_var = dataset.createVariable(variable_name, self.configs['variables'][variable_name]['data_type'])
                    temp_var[:] = self.configs['variables'][variable_name]['value']
                    
                    for key, value in variable_dict.items():
                        if (key != 'data_type') and (key != 'value'):
                            setattr(temp_var, key, value)


    def build(self):
        """
        Main loop - connects the reader with the writer.
        Iterates over input data files / file groups:
        - Reading meta attributes from "meta-data" configurations
        :return:
        """
        writer = None
        out_complete = ''

        reader = self.module_loader.get_reader()()
        reader.set_configs(self.configs)
        reader.verify_parameters()
        input_path = self.configs.get_resolved('parameters', 'input', 'path')
        output_path = self.configs.get_resolved('parameters', 'output', 'path')
        pathlib.Path(output_path).mkdir(parents=True, exist_ok=True)
        
        files = reader.fetch_input_files(input_path)

        for i, group in enumerate(files):

            obs = self.params('output_block_size') if self.configs.exists('parameters', 'output_block_size') else 1
            if obs is None:
                obs = len(files)

            first_of_batch = (i % obs == 0)

            if first_of_batch:
                output_name = reader.output_filename(group['id'])
                writer = self.module_loader.get_writer()(output_path, output_name)
                out_complete = writer.file_path()

            Logger.log('started_r_files', group['files'])

            with writer.appending(not first_of_batch) as dataset:
                Logger.log('writing_file', out_complete, '' if first_of_batch else '(appending)')
                
                self.read_attributes(dataset)
                self.read_variables(dataset)
                
                if reader.data_grouping:
                    complete_path = tuple([path.join(input_path, f) for f in group['files']])
                else:
                    complete_path = path.join(input_path, group['files'])
                reader.read_to(dataset, complete_path, self.configs, not first_of_batch)

        Logger.info('done')


def build(**args):
    builder = Builder(**args)
    builder.build()


import numpy as np
from pathlib import Path
from ..core.Reader import Reader
import datetime

class Windcubev2(Reader):

    def __init__(self):
        super().__init__(False)

    @staticmethod
    def util_process_time(timestring):
        hour = timestring[:2]
        minute = timestring[3:5]
        seconds = timestring[6:11]
        time = int(hour) * 3600
        time += int(minute) * 60
        time += float(seconds)
        return time

    @staticmethod
    def str_to_num(string1):
        try:
            return int(string1)
        except ValueError:
            try:
                return float(string1)
            except ValueError:
                return string1

    def accepts_file(self, filename):
        return filename.endswith(('.sta','.rtd','.stdsta'))

    def output_filename(self, filename):
        return filename[:-4]

    def read_to(self, output_dataset, input_filepath, configs, appending):
        try:
            # read file
            with open(input_filepath, encoding='latin-1') as f:
                data = f.readlines()
    
                # get parameters from header
                filetype = input_filepath[-3:]
                temp_HeaderSize = int(data[0].split(sep='=')[1])
                parameters = [line[:-1].split(sep='=') for line in data[0:temp_HeaderSize]]
                parameters = {line[0]: Windcubev2.str_to_num(line[1]) for line in parameters if len(line) == 2}
                parameters['Altitudes (m)'] = [Windcubev2.str_to_num(element) for element in parameters['Altitudes (m)'].strip().split('\t')] 
    
                
                if filetype == 'rtd':
                    parameters['first_timestamp'] = datetime.datetime.strptime(
                        data[parameters['HeaderSize'] + 2].split('\t')[0][:-3],
                        '%Y/%m/%d %H:%M:%S')
                
                else:
                    parameters['first_timestamp'] = datetime.datetime.strptime(
                        data[parameters['HeaderSize'] + 2].split('\t')[0],
                        '%Y/%m/%d %H:%M')
    
                # general data set description
                output_dataset.site = parameters['Location']
    
                # create the dimensions
                output_dataset.createDimension('range', len(parameters['Altitudes (m)']))
                output_dataset.createDimension('time', None)
    
                # create the coordinate variables
                range1 = output_dataset.createVariable('range', 'f4', ('range',))
                range1.units = 'm'
                range1.long_name = 'range_gate_distance_from_lidar'
                range1[:] = np.array(parameters['Altitudes (m)'])
    
                time = output_dataset.createVariable('time', str, ('time',))
                time.units = 's'
                time.long_name = 'Time UTC in ISO 8601 format yyyy-mm-ddThh:mm:ssZ'
    
                # create the beam steering and location variables
                yaw = output_dataset.createVariable('yaw', 'f4')
                yaw.units = 'degrees'
                yaw.long_name = 'lidar_yaw_angle'
                yaw[:] = parameters['DirectionOffset (°)']
    
                pitch = output_dataset.createVariable('pitch', 'f4')
                pitch.units = 'degrees'
                pitch.long_name = 'lidar_pitch_angle'
                pitch[:] = parameters['PitchAngle (°)']
    
                roll = output_dataset.createVariable('roll', 'f4')
                roll.units = 'degrees'
                roll.long_name = 'lidar_roll_angle'
                roll[:] = parameters['RollAngle (°)']
    
                # create the data variables
                scan_type = output_dataset.createVariable('scan_type', 'i')
                scan_type.units = 'none'
                scan_type.long_name = 'scan_type_of_the_measurement'
    
                scan_type[:] = 2
    
                accumulation_time = output_dataset.createVariable('accumulation_time', 'f4')
                accumulation_time.units = 'seconds'
                accumulation_time.long_name = 'time_for_spectral_accumulation'
                accumulation_time[:] = 1.0
    
                n_spectra = output_dataset.createVariable('n_spectra', 'f4')
                n_spectra.units = 'none'
                n_spectra.long_name = 'number_of_pulses'
                n_spectra[:] = parameters['Pulses / Line of Sight']
    
                # create the measurement variables
                if filetype == 'rtd':
                    VEL = output_dataset.createVariable('VEL', 'f4', ('time', 'range'))
                    VEL.units = 'm.s-1'
                    VEL.long_name = 'radial_velocity'
                    
                    azimuth_angle = output_dataset.createVariable('azimuth_angle', 'f4', ('time'))
                    azimuth_angle.units = 'degrees'
                    azimuth_angle.long_name = 'azimuth_angle_of_lidar beam'
                    
                    elevation_angle = output_dataset.createVariable('elevation_angle', 'f4', ('time'))
                    elevation_angle.units = 'degrees'
                    elevation_angle.long_name = 'elevation_angle_of_lidar beam'
                    				
                    T_internal = output_dataset.createVariable('T_internal', 'f4', ('time',))
                    T_internal.units = 'degrees C'
                    T_internal.long_name = 'internal_temperature'
                    
                    T_external = output_dataset.createVariable('T_external', 'f4', ('time',))
                    T_external.units = 'degrees C'
                    T_external.long_name = 'external_temperature'
                    
                    p = output_dataset.createVariable('p', 'f4', ('time',))
                    p.units = 'hPa'
                    p.long_name = 'pressure'
                    
                    Rh = output_dataset.createVariable('Rh', 'f4', ('time',))
                    Rh.units = 'percent'
                    Rh.long_name = 'relative_humidity'
    
                    CNR = output_dataset.createVariable('CNR', 'f4', ('time', 'range'))
                    CNR.units = 'dB'
                    CNR.long_name = 'carrier_to_noise_ratio'
                    				
                    WIDTH = output_dataset.createVariable('WIDTH', 'f4', ('time', 'range'))
                    WIDTH.units = 'm.s-1'
                    WIDTH.long_name = 'doppler_spectrum_width'
                                    
                    wiper = output_dataset.createVariable('wiper', 'f4', ('time',))
                    wiper.units = 'V'
                    wiper.long_name = 'Wiper count Vbatt'
                
                else:
                    WS = output_dataset.createVariable('WS', 'f4', ('time', 'range'))
                    WS.units = 'm.s-1'
                    WS.long_name = 'mean_of_scalar_wind_speed'
                    				
                    WSstd = output_dataset.createVariable('WSstd', 'f4', ('time', 'range'))
                    WSstd.units = 'm.s-1'
                    WSstd.long_name = 'standard_deviation_of_scalar_wind_speed'
                    				
                    WSmin = output_dataset.createVariable('WSmin', 'f4', ('time', 'range'))
                    WSmin.units = 'm.s-1'
                    WSmin.long_name = 'minimum_of_scalar_wind_speed'
                    				
                    WSmax = output_dataset.createVariable('WSmax', 'f4', ('time', 'range'))
                    WSmax.units = 'm.s-1'
                    WSmax.long_name = 'maximum_of_scalar_wind_speed'
                    
                    DIR = output_dataset.createVariable('DIR', 'f4', ('time', 'range'))
                    DIR.units = 'degrees'
                    DIR.long_name = 'mean_wind_direction'
                    
                    w = output_dataset.createVariable('w', 'f4', ('time', 'range'))
                    w.units = 'm.s-1'
                    w.long_name = 'mean_w_component_of_scalar_wind_speed'
                    				
                    wstd = output_dataset.createVariable('wstd', 'f4', ('time', 'range'))
                    wstd.units = 'm.s-1'
                    wstd.long_name = 'standard_deviation_of_w_component_of_scalar_wind_speed'
                    				
                    CNR = output_dataset.createVariable('CNR', 'f4', ('time', 'range'))
                    CNR.units = 'dB'
                    CNR.long_name = 'mean_carrier_to_noise_ratio'
                    
                    CNRmin = output_dataset.createVariable('CNRmin', 'f4', ('time', 'range'))
                    CNRmin.units = 'dB'
                    CNRmin.long_name = 'minimum_carrier_to_noise_ratio'
                    				
                    WIDTH = output_dataset.createVariable('WIDTH', 'f4', ('time', 'range'))
                    WIDTH.units = 'm.s-1'
                    WIDTH.long_name = 'mean_doppler_spectrum_width'
                    
                    Availability = output_dataset.createVariable('Availability', 'f4', ('time', 'range'))
                    Availability.units = 'percent'
                    Availability.long_name = 'data_availability'
                    				
                    T_internal = output_dataset.createVariable('T_internal', 'f4', ('time',))
                    T_internal.units = 'degrees C'
                    T_internal.long_name = 'mean_internal_temperature'
                    				
                    T_external = output_dataset.createVariable('T_external', 'f4', ('time',))
                    T_external.units = 'degrees C'
                    T_external.long_name = 'mean_external_temperature'
                    
                    p = output_dataset.createVariable('p', 'f4', ('time',))
                    p.units = 'hPa'
                    p.long_name = 'pressure'
                    
                    Rh = output_dataset.createVariable('Rh', 'f4', ('time',))
                    Rh.units = 'percent'
                    Rh.long_name = 'relative_humidity'	
                    
                    wiper = output_dataset.createVariable('wiper', 'f4', ('time',))
                    wiper.units = 'V'
                    wiper.long_name = 'Wiper count Vbatt'
    
                # fill values from dataset
                data_timeseries = [row.strip().split('\t') for row in data[parameters['HeaderSize'] + 2:]]
                
                if filetype == 'rtd': # high resolution data
                
                    timestamp_input = [datetime.datetime.strptime(row[0][:-3],'%Y/%m/%d %H:%M:%S') for row in data_timeseries]
                    timestamp_iso8601 = [value.isoformat()+'Z' for value in timestamp_input]
                    output_dataset.variables['time'][:] = np.array(timestamp_iso8601)
                    output_dataset.variables['T_internal'][:] = [float(row[2]) for row in data_timeseries]
                    output_dataset.variables['wiper'][:] = [float(row[3]) for row in data_timeseries]
                    output_dataset.variables['azimuth_angle'][:] = [float(row[1]) if row[1] != 'V' else 0 for row in
                                                                    data_timeseries]
                    output_dataset.variables['elevation_angle'][:] = [90 - parameters['ScanAngle (°)'] if row[1] != 'V' else 90
                                                                      for
                                                                      row
                                                                      in data_timeseries]
                    output_dataset.variables['VEL'][:, :] = [[float(value) for value in row[5::9]] for row in data_timeseries]
                    output_dataset.variables['WIDTH'][:, :] = [[float(value) for value in row[6::9]] for row in data_timeseries]
                    output_dataset.variables['CNR'][:, :] = [[float(value) for value in row[4::9]] for row in data_timeseries]
                    
    			# filetype == 'sta' # 10 minute mean values
                else:
                    timestamp_input = [datetime.datetime.strptime(row[0],'%Y/%m/%d %H:%M') for row in data_timeseries]
                    timestamp_iso8601 = [value.isoformat()+'Z' for value in timestamp_input]
                    output_dataset.variables['time'][:] = np.array(timestamp_iso8601)
                    output_dataset.variables['T_internal'][:] = [float(row[1]) for row in data_timeseries]
                    output_dataset.variables['T_external'][:] = [float(row[2]) for row in data_timeseries]
                    output_dataset.variables['p'][:] = [float(row[3]) for row in data_timeseries]
                    output_dataset.variables['Rh'][:] = [float(row[4]) for row in data_timeseries]				
                    output_dataset.variables['WS'][:, :] = [[float(value) for value in row[7::12]] for row in data_timeseries]
                    output_dataset.variables['WSstd'][:, :] = [[float(value) for value in row[8::12]] for row in data_timeseries]
                    output_dataset.variables['WSmin'][:, :] = [[float(value) for value in row[9::12]] for row in data_timeseries]
                    output_dataset.variables['WSmax'][:, :] = [[float(value) for value in row[10::12]] for row in data_timeseries]
                    output_dataset.variables['DIR'][:, :] = [[float(value) for value in row[11::12]] for row in data_timeseries]
                    output_dataset.variables['w'][:, :] = [[float(value) for value in row[12::12]] for row in data_timeseries]
                    output_dataset.variables['wstd'][:, :] = [[float(value) for value in row[13::12]] for row in data_timeseries]
                    output_dataset.variables['CNR'][:, :] = [[float(value) for value in row[14::12]] for row in data_timeseries]
                    output_dataset.variables['CNRmin'][:, :] = [[float(value) for value in row[15::12]] for row in data_timeseries]				
                    output_dataset.variables['WIDTH'][:, :] = [[float(value) for value in row[16::12]] for row in data_timeseries]
                    output_dataset.variables['Availability'][:, :] = [[float(value) for value in row[17::12]] for row in data_timeseries]
                    output_dataset.variables['wiper'][:] = [float(row[6]) for row in data_timeseries]

        except Exception as err:
            print('Error ocurred while converting %s. See error.log for details.' % input_filepath)
       
            with open(Path(output_dataset.filepath()).parent / 'error.log','a') as logfile:
                logfile.write( '%s'%output_dataset.filepath() +'\n')



from abc import ABC, abstractmethod
from os import path


class Writer(ABC):
    """
    Specifies a writer class API and implements some common methods.
    """

    dir_path = None
    name = None

    def __init__(self, dir_path, name):
        """
        Constructor.
        :param dir_path: directory to save the file into.
        :param name: file name suggestion (without extension)
        :param append: should be the data appended or a new file writen
        """
        super().__init__()
        self.dir_path = dir_path
        self.name = name
        self.append = False

    def file_path(self):
        """
        Complete file (to be written to) path.
        :return: file path
        """
        return path.join(self.dir_path, self.filename())

    @abstractmethod
    def filename(self):
        """
        Returns the filename (with extension) of the file to be writen.
        :return: filename
        """
        pass

    def appending(self, append):
        """
        Sets the writer appending mode.
        :param append: boolean
        :return: the writer itself.
        """
        self.append = append
        return self

    @abstractmethod
    def __enter__(self):
        """
        Requests an CDM object reference that is used to represent the dataset file
        :return: the dataset object
        """
        pass

    @abstractmethod
    def __exit__(self, type, value, traceback):
        """
        Closes/Writes the dataset
        :param type:
        :param value:
        :param traceback:
        :return:
        """
        pass



import netCDF4 as  nc

from ..core.Writer import Writer


class NetCDF4(Writer):
    dataset = None

    def __init__(self, dir_path, name):
        super().__init__(dir_path, name)

    def filename(self):
        return self.name + '.nc'

    def __enter__(self):
        self.dataset = nc.Dataset(self.file_path(), 'a' if self.append else 'w', format='NETCDF4')
        return self.dataset.__enter__()

    def __exit__(self, type, value, traceback):
        return self.dataset.__exit__(type, value, traceback)



from .core.Builder import Builder
from .common.Logger import Logger

from os import path
import argparse


def main():
    """
        Converter command entry point.
    """

    parser = argparse.ArgumentParser()

    parser.add_argument('-C', '--config-file', default='config.yaml',
                        help='Configuration file path (default: configs.xml)')
    parser.add_argument('-O', '--output-format', default=None,
                        help='Output format produced: NetCDF4, HDF5, metainfo,...')
    parser.add_argument('-I', '--input-format', default=None,
                        help='Input files format as produced by the Lidar: S100, V1,...')
    parser.add_argument('-D', '--input-path', default=None,
                        help='Input datasets directory path')
    parser.add_argument('-v', '--verbose', action='store_true', default=False,
                        help='explain what is being done')
    parser.add_argument('-V', '--version', action='store_true', default=False,
                        help='display version information and exit')
    parser.add_argument('--debug', action='store_true', default=False,
                        help='Shows internal error messages')
    parser.add_argument('--context', default='',
                        help='Path to which relative paths, such as config-file or data-path, are resolved.')

    args = parser.parse_args()

    Logger.set_args(args)
    Logger.header()

    if not args.version:
        args_dict = vars(args)
        args_dict.pop('verbose')
        args_dict.pop('version')
        args_dict.pop('debug')
        builder = Builder(**args_dict)
        builder.build()
    else:
        Logger.log('about')



from ..core.Reader import Reader
import numpy as np
import datetime

# Change this to parameters, if necessary.
doppler_index = 1
intensity_index = 2
time_index = 3
azimuth_index = 4
elevation_index = 5
pitch_index = 6
roll_index = 7


def chunks(l, n):
    for i in range(0, len(l), n):
        yield l[i:i + n]


def process_time(timestamp):
    y, m, d = timestamp[:10].split('-')
    h, min_, s = timestamp[11:19].split(':')
    ms = timestamp[20:]
    dt = datetime.datetime(int(y), int(m), int(d), int(h), int(min_), int(s), int(ms) * 1000)
    return dt.timestamp()


def create_variables(group, azimuth, elevation, _yaw, _pitch, _roll, doppler, intensity):
    # create the beam steering variables
    # azimuth and elevation
    azimuth_angle = group.createVariable('azimuth_angle', 'f4', 'time')
    azimuth_angle.units = 'degrees'
    azimuth_angle.long_name = 'azimuth_angle_of_lidar beam'
    azimuth_angle[:] = azimuth
    azimuth_angle.comment = ''
    azimuth_angle.accuracy = ''
    azimuth_angle.accuracy_info = ''

    elevation_angle = group.createVariable('elevation_angle', 'f4', 'time')
    elevation_angle.units = 'degrees'
    elevation_angle.long_name = 'elevation_angle_of_lidar beam'
    elevation_angle[:] = elevation
    elevation_angle.comment = ''
    elevation_angle.accuracy = ''
    elevation_angle.accuracy_info = ''

    # yaw, pitch, roll
    yaw = group.createVariable('yaw', 'f4', 'time')
    yaw.units = 'degrees'
    yaw.long_name = 'lidar_yaw_angle'
    yaw[:] = _yaw
    yaw.comment = 'The home position of the lidar has been configured in a way that 0 azimuth corresponds to ' \
                  'north. '
    yaw.accuracy = ''

    pitch = group.createVariable('pitch', 'f4', 'time')
    pitch.units = 'degrees'
    pitch.long_name = 'lidar_pitch_angle'
    pitch[:] = _pitch
    pitch.comment = ''
    pitch.accuracy = ''
    pitch.accuracy_info = 'No information on pitch accuracy available.'

    roll = group.createVariable('roll', 'f4', 'time')
    roll.units = 'degrees'
    roll.long_name = 'lidar_roll_angle'
    roll[:] = _roll
    roll.comment = ''
    roll.accuracy = ''
    roll.accuracy_info = 'No information on roll accuracy available.'

    # measurement variables
    # TODO verify metadata
    # Doppler and Intensity
    DOPPLER = group.createVariable('DOPPLER', 'f4', ('time', 'range'))
    DOPPLER.units = 'm.s-1'
    DOPPLER.long_name = 'doppler'
    DOPPLER[:] = doppler

    INTENSITY = group.createVariable('INTENSITY', 'f4', ('time', 'range'))
    INTENSITY.units = 'm.s-1'
    INTENSITY.long_name = 'intensity'
    INTENSITY[:] = intensity


class Galion(Reader):

    def __init__(self):
        super().__init__(False)

    def accepts_file(self, filename):
        return filename.endswith('.scn')

    def output_filename(self, filename):
        return filename[:-4]

    def required_params(self):
        return ['n_gates', 'range_gates', 'constant_gates', 'measurement_scenarios']

    def read_to(self, output_dataset, input_filepath, configs, appending):
        with open(input_filepath) as file:
            nr_gates = configs['parameters']['n_gates']
            range_gates = configs['parameters']['range_gates']
            constant_gates = configs['parameters']['constant_gates']
            measurement_scenarios = configs['parameters']['measurement_scenarios']
            raw_file = file.readlines()

            # if we can assume that the metadata is always the same length
            # If not, then it's better to add a variable to identify the number of lines that are "metadata"
            metadata = raw_file[:6]
            data = raw_file[6:]

            scans = np.array(list(chunks([row.strip().split('\t') for row in data], int(nr_gates))))

            # create the dimensions
            output_dataset.createDimension('range', nr_gates)
            output_dataset.createDimension('time', len(scans))

            # create the coordinate variables
            # range
            range1 = output_dataset.createVariable('range', 'f4', ('range',))
            range1.units = 'm'
            range1.long_name = 'range_gate_distance_from_lidar'
            if constant_gates:
                range1[:] = np.full(nr_gates, float(range_gates))
            else:
                range1[:] = np.array(range_gates.split(';')).astype(float)
            range1.comment = ''

            # time
            time = output_dataset.createVariable('time', 'f4', ('time',))
            time.units = 's'
            start_time_kv = metadata[4]
            start_time_str = start_time_kv[start_time_kv.find('\t') + 1:start_time_kv.find('\n')]
            start_time_seconds = process_time(start_time_str)
            time.long_name = 'seconds since ' + start_time_str
            # array manipulation to obtain timestamps
            timestamps = (scans[:, :, time_index])[:, 0]
            time[:] = np.array(list(map(lambda x: process_time(x) - start_time_seconds, timestamps)))
            time.comment = ''

            # create the data variables
            scan_type = output_dataset.createVariable('scan_type', 'i', 'time')
            scan_type.units = 'none'
            scan_type.long_name = 'scan_type_of_the_measurement'

            scan_id = output_dataset.createVariable('scan_id', 'i', 'time')
            scan_id.units = 'none'
            scan_id.long_name = 'scan_id_of_the_measurement'

            create_variables(output_dataset, (scans[:, :, azimuth_index])[:, 0], (scans[:, :, elevation_index])[:, 0], np.zeros(len(scans)),
                             (scans[:, :, pitch_index])[:, 0], (scans[:, :, roll_index])[:, 0],
                             (scans[:, :, doppler_index]), (scans[:, :, intensity_index]))

            invalid_scans = 0
            scan_index = 1
            for s in measurement_scenarios:
                long_name = s['scenario']
                _type = int(s['type'])
                records = s['scans']
                if long_name == 'INVALID':
                    split_scans = records.split(';')
                    for ss in split_scans:
                        invalid_scans += int(ss.split('-')[1]) - int(ss.split('-')[0]) + 1
                else:
                    scan_group = output_dataset.createGroup('scan_' + str(scan_index) + '_' + long_name)
                    _azimuth = np.zeros(len(scans))
                    _elevation = np.zeros(len(scans))
                    _yaw = np.zeros(len(scans))
                    _pitch = np.zeros(len(scans))
                    _roll = np.zeros(len(scans))
                    _doppler = np.zeros(shape=[len(scans), int(nr_gates)])
                    _intensity = np.zeros(shape=[len(scans), int(nr_gates)])
                    split_scans = records.split(';')
                    for ss in split_scans:
                        initial_index = int(ss.split('-')[0]) - (invalid_scans + 1)
                        final_index = int(ss.split('-')[1]) - (invalid_scans + 1)
                        scan_type[initial_index:final_index + 1] = _type
                        scan_id[initial_index:final_index + 1] = scan_index
                        _azimuth[initial_index:final_index + 1] = scans[initial_index:final_index + 1, :, azimuth_index][:, 0]
                        _elevation[initial_index:final_index + 1] = scans[initial_index:final_index + 1, :, elevation_index][:, 0]
                        _pitch[initial_index:final_index + 1] = scans[initial_index:final_index + 1, :, pitch_index][:, 0]
                        _roll[initial_index:final_index + 1] = scans[initial_index:final_index + 1, :, roll_index][:, 0]
                        _doppler[initial_index:final_index + 1] = scans[initial_index:final_index + 1, :, doppler_index]
                        _intensity[initial_index:final_index + 1] = scans[initial_index:final_index + 1, :, intensity_index]
                    create_variables(scan_group, _azimuth, _elevation, _yaw, _pitch, _roll, _doppler, _intensity)
                    scan_index += 1



from ..core.Reader import Reader
from datetime import datetime, timedelta
import numpy as np
import re

class Triton(Reader):

    def __init__(self):
        super().__init__(False)

    def accepts_file(self, filename):
        return filename.startswith('TritonExport') & (len(filename) > 14) & filename.endswith('.csv')

    def output_filename(self, timestamp):
        return timestamp[:-9]

    def read_to(self, output_dataset, input_filepaths, parameters, appending):
        wind_file = input_filepaths        
        
        with open(wind_file) as f:
            wind_file_data = f.readlines()

        wind_file_data = [row.strip().split(';') for row in wind_file_data]
        wind_file_data_T = list(zip(*wind_file_data[4:]))


        if not appending:

            
            range_list=[]
            for column in wind_file_data[2]:
                temp = re.findall('\d+(?=m)',column)
                if len(temp) > 0:
                    range_list.append(int(temp[0]))

            range_list = list(set(range_list))
            range_list.sort()
            
            # create the dimensions
            output_dataset.createDimension('range', len(range_list))
            output_dataset.createDimension('time', None)

            # create the coordinate variables

            # range
            range1 = output_dataset.createVariable('range', 'f4', ('range',))
            range1.units = 'm'
            range1.long_name = 'range_gate_distance_from_lidar'
            range1[:] = range_list
            range1.comment = ''

            # time
            time = output_dataset.createVariable('time', str, ('time',))
            time.units = 's'
            time.long_name = 'seconds since 1904-01-01 12:00AM UTC'
            time.comment = ''

            # create the data variables
            scan_type = output_dataset.createVariable('scan_type', 'i')
            scan_type.units = 'none'
            scan_type.long_name = 'scan_type_of_the_measurement'


            # create the measurement variables VEL, Quality, WIDTH
            VEL = output_dataset.createVariable('VEL', 'f4', ('time', 'range'))
            VEL.units = 'm.s-1'
            VEL.long_name = 'radial velocity'
            VEL.comment = ''
            VEL.accuracy = ''
            VEL.accuracy_info = ''
            
            DIR = output_dataset.createVariable('DIR', 'f4', ('time', 'range'))
            DIR.units = 'degrees north'
            DIR.long_name = 'wind direction from north'
            
            
            Quality = output_dataset.createVariable('Quality', 'f4', ('time', 'range'))
            Quality.units = 'percent'
            Quality.long_name = 'quality_value'
            Quality.comment = ''
            Quality.accuracy = ''
            Quality.accuracy_info = ''

            w = output_dataset.createVariable('w', 'f4', ('time', 'range'))
            w.units = 'm.s-1'
            w.long_name = 'vertical_wind_speed'
            w.comment = ''
            w.accuracy = ''
            w.accuracy_info = ''
            
            #%% get timestamps in ISO 8601 format

            timestamp_list = [datetime.strptime(value, '%d.%m.%Y %H:%M') for value in wind_file_data_T[0]]
            timestamp_iso8601 = [value.isoformat()+'Z' for value in timestamp_list]
            output_dataset.variables['time'][:] = np.array(timestamp_iso8601)

            #%% read vel, width, Quality out of dataset
            # e.g. radial velocity starts at 5th column and is then repeated every 9th column
            
            def str2float(astring):
                if len(astring) > 0:
                    return float(astring.replace(',','.'))
                else:
                    return 0

            
            output_dataset.variables['VEL'][:, :] = list(
                zip(*[[str2float(value) for value in row] for row in wind_file_data_T[2:len(range_list)*4+1:4]]))
            output_dataset.variables['DIR'][:, :] = list(
                zip(*[[str2float(value) for value in row] for row in wind_file_data_T[1:len(range_list)*4+1:4]]))
            output_dataset.variables['w'][:, :] = list(
                zip(*[[str2float(value) for value in row] for row in wind_file_data_T[3:len(range_list)*4+1:4]]))
            output_dataset.variables['Quality'][:, :] = list(
                zip(*[[str2float(value) for value in row] for row in wind_file_data_T[4:len(range_list)*4+1:4]]))
            
        #%% case appending
        else: 
            ntime = len(output_dataset.dimensions["time"])
            
            timestamp_list = [datetime.strptime(value, '%d.%m.%Y %H:%M') for value in wind_file_data_T[0]]
            timestamp_iso8601 = [value.isoformat()+'Z' for value in timestamp_list]
            output_dataset.variables['time'][ntime:] = np.array(timestamp_iso8601)

            # e.g. radial velocity starts at 5th column and is then repeated every 9th column
            output_dataset.variables['VEL'][ntime:, :] = list(
                zip(*[[str2float(value) for value in row] for row in wind_file_data_T[2:len(range_list)*4+1:4]]))
            output_dataset.variables['DIR'][ntime:, :] = list(
                zip(*[[str2float(value) for value in row] for row in wind_file_data_T[1:len(range_list)*4+1:4]]))
            output_dataset.variables['w'][ntime:, :] = list(
                zip(*[[str2float(value) for value in row] for row in wind_file_data_T[3:len(range_list)*4+1:4]]))
            output_dataset.variables['Quality'][ntime:, :] = list(
                zip(*[[str2float(value) for value in row] for row in wind_file_data_T[4:len(range_list)*4+1:4]]))



import netCDF4 as nc
import json

from ..core.Writer import Writer


class MetadataCard(Writer):
    nc_dataset = None

    def __init__(self, dir_path, name):
        super().__init__(dir_path, name)

    def filename(self):
        return self.name + '.json'

    def __enter__(self):
        if not self.append:
            self.nc_dataset = nc.Dataset(self.file_path(), 'w', diskless=True)

        return self.nc_dataset.__enter__()

    def __exit__(self, type, value, traceback):
        if not self.append:

            metadata_card = {}

            # Writing attributes
            for ncattr in self.nc_dataset.ncattrs():
                metadata_card[ncattr] = self.nc_dataset.getncattr(ncattr)

            with open(self.file_path(), "w") as json_file:
                json_file.write(json.dumps(metadata_card, indent=4))



import importlib


class ModuleLoader:
    """
    Dynamically handles "add-on" modules (readers and writers) loading.
    """

    def __init__(self):
        super().__init__()
        self.reader_module = None
        self.writer_module = None

    @staticmethod
    def load(path, name):
        """
        Dynamically loads a module with name "name" and retrieves
        the class with the same name declared in that file.
        :param path: should be 'reader' or 'writer'
        :param name: worker name e.g. windscanner, windcubev2, netcdf4 etc.
        Check out the available ones at readers/writers sub directories.
        :return: loaded class.
        """
        return getattr(importlib.import_module(path + name, __package__), name)

    def load_reader(self, name):
        """
        Loads a reader and stores it into reader_modules dictionary.
        :param name: reader name
        :return: void
        """
        self.reader_module = self.load('..readers.', name)

    def load_writer(self, name):
        """
        Loads a writer and stores it into writer_modules dictionary.
        :param name: writer name
        :return: void
        """
        self.writer_module = self.load('..writers.', name)

    def get_reader(self):
        """
        Returns the reader class.
        :return: class reference
        """
        return self.reader_module

    def get_writer(self):
        """
        Returns the writer class.
        :return: class reference
        """
        return self.writer_module


    def set_reader(self, reader):
        self.reader_module = reader

    def set_writer(self, writer):
        self.writer_module = writer



from abc import ABC, abstractmethod
from os import listdir
from itertools import groupby
from lidaco.common.Logger import Logger


class Reader(ABC):
    """
    Specifies a reader class API and implements some common methods.
    """

    out_path = ''

    def __init__(self, data_grouping):
        """
        Constructor.
        :param data_grouping: If data_grouping is True then group_id method must be implemented.
        """
        super().__init__()
        self.data_grouping = data_grouping
        self.configs = None

    def fetch_input_files(self, dir_path):
        """
        Lists and filters input data files. If the reader specifies a group_by function,
        it also groups files by that value, return a dictionary group => [files].
        :param dir_path: directory containing input data files
        :return: [files] | {group => [files]}
        """
        Logger.info('searching_in_path', dir_path)
        files = [f for f in listdir(dir_path) if self.accepts_file(f)]
        [Logger.info('found', f) for f in files]

        if len(files) == 0:
            Logger.error('files_not_found')

        if self.data_grouping:
            Logger.info('grouping')
            files.sort()
            input_files = []

            for k, g in groupby(files, self.group_id):
                input_files.append({'id': k, 'files': list(g)})
        else:
            input_files = [{'id': f, 'files': f} for f in files]

        input_files.sort(key=lambda group: group['id'])
        return input_files

    @abstractmethod
    def accepts_file(self, filename):
        """
        Checks if a file is valid. Used by the converter to filter applicable files.
        :param filename: e.g., 20161211135000_wind.txt, WLS7-164_2016_09_19__00_00_00.rtd
        :return: boolean
        """
        pass

    @abstractmethod
    def output_filename(self, input_filename):
        """
        Checks if a file is valid. Used by the converter to filter applicable files.
        :param input_filename: input filename or group name when group_by is used.
        When output_block_size is set, the output_filename is set wrt the first file/group's name
        :return: output filename (without extension)
        """
        pass

    def required_params(self):
        """
        Specifies what parameters the user must specify in the configuration files.
        It should be overriten by the reader implementation when needed.
        The conversion will halt if any of these are not specified.
        :return: a list with the parameters names
        """
        return []

    def verify_parameters(self):
        def verify_block(block, path):
                if type(block) == dict :
                    for k, value in block.items():
                        verify_block(value, path + [k])
                elif type(block) == list:
                    for value in block:
                        verify_block(value, path)
                else:
                    args = path + [block]
                    try:
                        self.configs.get(*args)
                    except Exception as e:
                        Logger.error('missing_reader_param', '/'.join(args), type(self).__name__)
        verify_block(self.required_params(), ['parameters'])

    @abstractmethod
    def read_to(self, output_dataset, input, configurations, index):
        """
        Reads an input file/group to the correspondent output dataset.
        :param output_dataset: cdm/netcdf4 dataset.
        :param input: is file path.
        When group_by is used, a tuple containing the group file paths.
        :param configurations: configurations read from .yaml files
        :param index:
        :return: void
        """
        pass

    def group_id(self, filename):
        """
        Used by the converter to group by the converter to combine multiple files into a group.
        :param filename: e.g., 20161211135000_wind.txt
        :return: group name e.g. 20161211135000
        """
        pass

    def config(self, *keys):
        return self.configs.get('parameters', *keys)

    def set_configs(self, configs):
        self.configs = configs



from ..core.Reader import Reader
from datetime import datetime
import numpy as np


class WLS70(Reader):

    def __init__(self):
        super().__init__(False)

    def accepts_file(self, filename):
        return (filename[-4:] == '.txt') & (len(filename) > 14) & (filename[:5] == 'WLS70')

    def output_filename(self, timestamp):
        return timestamp[:-9]
    
    def read_to(self, output_dataset, input_filepaths, parameters, appending):
        wind_file = input_filepaths

        with open(wind_file) as f:
            input_file_data = [line.strip() for line in f.readlines()]
            

        
        ordered_data=np.reshape(np.array(input_file_data)[39:-1],(-1,17))
        range_list = np.unique(ordered_data[:,9].astype(float)).astype(int)
        
        time_array = ordered_data[::len(range_list),1:7].astype(int)
        datetime_array = [datetime(*line) for line in time_array]
        iso8601_array = [date.isoformat()+'Z' for date in datetime_array]


        if not appending:

            output_dataset.createDimension('range', len(range_list))
            output_dataset.createDimension('time', None)

            time = output_dataset.createVariable('time', str, ('time',))
            time.units = 's'
            time.long_name = 'Time UTC in ISO 8601 format yyyy-mm-ddThh:mm:ssZ'
            time.comment = ''
            time[:] = np.array(iso8601_array)

            T_internal = output_dataset.createVariable('T_internal', 'f4', ('time','range'))
            T_internal.units = 'degrees C'
            T_internal.long_name = 'temperature'
            T_internal[:] = np.reshape(ordered_data[:,7].astype(float),(len(iso8601_array),len(range_list)))
            
            elevation_angle = output_dataset.createVariable('elevation_angle', 'f4', ('time','range'))
            elevation_angle.units = 'degrees'
            elevation_angle.long_name = 'elevation_angle_of_lidar beam'
            elevation_angle[:] = np.reshape(ordered_data[:,8].astype(float),(len(iso8601_array),len(range_list)))

            range1 = output_dataset.createVariable('range', 'f4', ('range',))
            range1.units = 'm'
            range1.long_name = 'range_gate_distance_from_lidar'
            range1.comment = ''
            range1[:] = range_list
            
            CNR = output_dataset.createVariable('CNR', 'f4', ('time', 'range'))
            CNR.units = 'dB'
            CNR.long_name = 'carrier-to-noise ratio'
            CNR.comment = ''
            CNR.accuracy = ''
            CNR.accuracy_info = ''
            CNR[:] = np.reshape(ordered_data[:,10].astype(float),(len(iso8601_array),len(range_list)))
            
            VEL = output_dataset.createVariable('VEL', 'f4', ('time', 'range'))
            VEL.units = 'm.s-1'
            VEL.long_name = 'radial velocity'
            VEL.comment = ''
            VEL.accuracy = ''
            VEL.accuracy_info = ''
            VEL[:] = np.reshape(ordered_data[:,11].astype(float),(len(iso8601_array),len(range_list)))
            
            DIR = output_dataset.createVariable('DIR', 'f4', ('time', 'range'))
            DIR.units = 'degrees north'
            DIR.long_name = 'wind direction from north'
            DIR[:] = np.reshape(ordered_data[:,12].astype(float),(len(iso8601_array),len(range_list)))
            
        else: 
            ntime = len(output_dataset.dimensions["time"])
            output_dataset.variables['time'][ntime:] = np.array(iso8601_array)
            output_dataset.variables['T_internal'][ntime:] = np.reshape(ordered_data[:,7].astype(float) , (len(iso8601_array),len(range_list)))
            output_dataset.variables['elevation_angle'][ntime:] = np.reshape(ordered_data[:,8].astype(float) , (len(iso8601_array),len(range_list)))
            output_dataset.variables['CNR'][ntime:] = np.reshape(ordered_data[:,10].astype(float) , (len(iso8601_array),len(range_list)))
            output_dataset.variables['VEL'][ntime:] = np.reshape(ordered_data[:,11].astype(float)  ,(len(iso8601_array),len(range_list)))        
            output_dataset.variables['DIR'][ntime:] = np.reshape(ordered_data[:,12].astype(float),(len(iso8601_array),len(range_list)))


from ..common.Utils import dict_merge, map_recursively
from ..common.Logger import Logger
from yaml import load
from os import path


class Config:
    """
    Class responsible to handle hierarchical configurations.
    Can be used as a regular dictionary.

    """

    def __init__(self, context, file_name=None, configs={}):
        """
        Loads a configuration file and the declared imports in it recursively.
        :param context: file location path
        :param file_name: configuration filename
        """
        self.configs = {}
        self.config_paths = {}
        self.context = context

        tmp_configs = {}

        if file_name:
            tmp_configs = self.load_from_file(file_name)

        dict_merge(tmp_configs, configs)  # apply argument passed configs

        if 'imports' in tmp_configs:
            self.resolve_imports(context, tmp_configs.pop('imports'))

        dict_merge(self.configs, tmp_configs)
        dict_merge(self.config_paths, map_recursively(tmp_configs, context))

    def resolve_imports(self, dir_path, imports):
        """

        :param dir_path:
        :param imports:
        :return:
        """
        for relative_path in imports:
            absolute_path = path.join(dir_path, relative_path)
            import_dir_path = path.dirname(absolute_path)
            import_filename = path.basename(absolute_path)

            config = Config(import_dir_path, import_filename)
            dict_merge(self.configs, config.get())
            dict_merge(self.config_paths, config.get_path())

    def load_from_file(self, file_name):
        full_path = ""
        try:
            full_path = path.join(self.context, file_name)
            Logger.info('loading_config', full_path.replace("/./", "/"))
            with open(full_path, 'r') as stream:
                return load(stream)
        except FileNotFoundError as e:
            Logger.error('bad_config_path', full_path.replace("/./", "/"))
        except Exception as e:
            Logger.error('bad_config_formatting', str(e))

    def merge(self, config):
        """
        Updates the current configuration data with "config"
        Overriding/Update is performed at the leaf level.
        :param config: other config structure
        :return:
        """
        dict_merge(self.configs, config)

    def get(self, *keys):
        keys = list(keys)
        tmp_value = self.configs
        while len(keys) > 0:
            tmp_value = tmp_value[keys[0]]
            keys.pop(0)

        return tmp_value

    def get_path(self, *keys):
        keys = list(keys)
        tmp_value = self.config_paths
        while len(keys) > 0:
            tmp_value = tmp_value[keys[0]]
            keys.pop(0)

        return tmp_value

    def exists(self, *keys):
        try:
            self.get(*keys)
            return True
        except KeyError:
            return False

    def __getitem__(self, key):
        """
        Returns an item from the config data.
        :param key: key name
        :return: value
        """
        return self.configs[key]

    def __contains__(self, key):
        """
        Checks if a certain key exists in the config data.
        :param key: key name
        :return: boolean
        """
        return key in self.configs

    def get_resolved(self, *key):
        return path.join(self.get_path(*key), self.get(*key))



from setuptools import setup, find_packages


# read the contents of your README file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


setup(
    name='lidaco',
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=find_packages(),
    version='0.0.21',
    description='Wind Lidar Data Converter.',
    author='e-WindLidar',
    entry_points={
        'console_scripts': [
            'lidaco = lidaco.__main__:main'
        ],
    },
    # scripts=['bin/lidaco'],
    author_email='e-windlidar@googlegroups.com',
    url='https://github.com/e-WindLidar/Lidaco',  # use the URL to the github repo
    # download_url='https://github.com/e-WindLidar/Lidaco/archive/v0.0.1.tar.gz',
    keywords=['wind', 'lidar', 'data', 'converter'],  # arbitrary keywords
    classifiers=[],
    install_requires=[
        'pyyaml',
        'netCDF4',
        'lxml'
    ]
)



import numpy as np
from ..core.Reader import Reader
import datetime
import pandas as pd
import re
from pathlib import Path

class ZephIR300(Reader):

    def __init__(self):
        super().__init__(False)

    def accepts_file(self, filename):
        return (filename.endswith(('.csv','.CSV')) & filename.startswith('Wind'))

    def output_filename(self, filename):
        return filename[:-4]

    def parse_time(string1):
            try:
                temp = datetime.datetime.strptime(string1,'%d.%m.%Y %H:%M:%S')
            except:
                try:
                    temp = datetime.datetime.strptime(string1,'%d/%m/%Y %H:%M:%S')
                except:
                    temp = datetime.datetime.strptime(string1,'%d.%m.%Y %H:%M')
            temp = temp.isoformat() +'Z'
            return temp

#    def required_params(self):
#        return ['position_x_input', 'position_y_input', 'position_z_input']

    def read_to(self, output_dataset, input_filepath, configs, appending):
       
        # read file
        
        ten_min_file = (re.findall(r'(?<=\\)\w+(?=_\d+@)',input_filepath)[0] == r'Wind10')
#
        try:
            f = open(input_filepath)
            f.readline()
            header = f.readline()
            header = header.split('Checksum')
            myCols = header[0].split(';')
            df=pd.read_csv(input_filepath,sep=';',skiprows=1,decimal=',',usecols=range(len(myCols)),index_col=False)   
            
    
            with open(input_filepath,'r',encoding='latin-1') as f: # get parameters from header
                header = f.readline()
                header=header.split(';')
                parameters = {line.split(':')[0].strip(): line.split(':')[1].strip() for line in header if ':' in line}
                parameters['Measurement heights'] = [int(element.strip()) for element in parameters['Measurement heights'].split('m') if (element != '')]
                parameters['Measurement heights'].append(1)
    
    
            # create the dimensions
            output_dataset.createDimension('range', len(parameters['Measurement heights']))
            output_dataset.createDimension('time', None)
    
            # create the coordinate variables
            range1 = output_dataset.createVariable('range', 'f4', ('range',))
            range1.units = 'm'
            range1.long_name = 'range_gate_distance_from_lidar'
            range1[:] = np.array(parameters['Measurement heights'])
    
            time = output_dataset.createVariable('time', str, ('time',))
            time.units = 's'
            time.long_name = 'timestamp ISO 8601'
    
            # create the data variables
            scan_type = output_dataset.createVariable('scan_type', 'i')
            scan_type.units = 'none'
            scan_type.long_name = 'scan_type_of_the_measurement'
            scan_type[:] = 1
    
            accumulation_time = output_dataset.createVariable('accumulation_time', 'f4')
            accumulation_time.units = 'seconds'
            accumulation_time.long_name = 'time_for_spectral_accumulation'
            accumulation_time[:] = 1.0
    
            # create the measurement variables
            
            # Beschreibung einfügen
            tilt = output_dataset.createVariable('tilt', 'f4', ('time',))
            tilt.units = 'degrees north'
            tilt.long_name = 'either pitch or roll depending on higher value'
    
            T_external = output_dataset.createVariable('T_external', 'f4', ('time',))
            T_external.units = 'degrees C'
            T_external.long_name = 'temperature'
    
            yaw = output_dataset.createVariable('yaw', 'f4', ('time',))
            yaw.units = 'degrees'
            yaw.long_name = 'lidar_yaw_angle'
    
            rh = output_dataset.createVariable('rh', 'f4', ('time',))
            rh.units = 'degrees'
            rh.long_name = 'lidar_yaw_angle'
            
            p = output_dataset.createVariable('p', 'f4', ('time',))
            p.units = 'degrees'
            p.long_name = 'lidar_yaw_angle'
            
            if ten_min_file:
                n_valid = output_dataset.createVariable('n_valid', 'f4', ('time', 'range'))
                n_valid.units = '-'
                n_valid.long_name = 'number of valid scans in averaging period'
    
                if 'Proportion Of Packets With Rain (%)' in df.columns:
                    proportion_of_rain = output_dataset.createVariable('proportion_of_rain', 'f4', ('time',))
                    proportion_of_rain.units = 'percent'
                    proportion_of_rain.long_name = 'Proportion Of Packets With Rain'
                elif 'Raining' in df.columns:
                    proportion_of_rain = output_dataset.createVariable('rain', 'f4', ('time',))
                    proportion_of_rain.units = 'boolean'
                    proportion_of_rain.long_name = 'indictor for rain; 1 is rain 0 no rain'
                

            WS = output_dataset.createVariable('WS', 'f4', ('time', 'range'))
            WS.units = 'm.s-1'
            WS.long_name = 'mean of scalar wind speed'
            
            DIR = output_dataset.createVariable('DIR', 'f4', ('time', 'range'))
            DIR.units = 'degrees north'
            DIR.long_name = 'wind direction from north'
    
    
    
    
    
            # fill values from dataset        
            df['timestamp_iso8601'] = df['Time and Date'].apply(ZephIR300.parse_time)
        
    
            output_dataset.variables['time'][:] = df['timestamp_iso8601'].values        
            output_dataset.variables['T_external'][:] = df['Air Temp. (C)'].values        
            output_dataset.variables['tilt'][:] = df['Tilt (deg)'].values
            output_dataset.variables['yaw'][:] = df['ZephIR Bearing (deg)'].values
            output_dataset.variables['rh'][:] = df['Humidity (%)'].values
            output_dataset.variables['p'][:] = df['Pressure (mbar)'].values
            
            if ten_min_file:
                if 'Proportion Of Packets With Rain (%)' in df.columns:
                    output_dataset.variables['proportion_of_rain'][:] = df['Proportion Of Packets With Rain (%)'].values
                elif 'Raining' in df.columns:
                    output_dataset.variables['rain'][:] = df['Raining'].values
            
            met_ws_list = df.iloc[:,16]
            met_dir_list = df.iloc[:,17]
    
            if ten_min_file:
                ws_list = df.iloc[:,21:-2:8]
                dir_list = df.iloc[:,20:-2:8]
                n_valid_list = df.iloc[:,19:-2:8]
                n_valid_complete = pd.concat([n_valid_list,pd.Series(np.full_like(met_ws_list, np.nan),name='Packets in Average at MET')],join='inner',axis=1)
                output_dataset.variables['n_valid'][:, :] = n_valid_complete.values
    
            else:
                ws_list = df.iloc[:,20:-2:3]
                dir_list = df.iloc[:,19:-2:3]  
            
            ws_list_complete = pd.concat([ws_list, met_ws_list],join='inner',axis=1)
            dir_list_complete = pd.concat([dir_list, met_dir_list],join='inner',axis=1)
    
            output_dataset.variables['WS'][:, :] = ws_list_complete.values
            output_dataset.variables['DIR'][:, :] = dir_list_complete.values

        except Exception as err:
                print('Error ocurred while converting %s. See error.log for details.' % input_filepath)
                print(err)
           
                with open(Path(output_dataset.filepath()).parent / 'error.log','a') as logfile:
                    logfile.write( '%s'%output_dataset.filepath() +'\n')

